/*******************************************************************
* DivXripper - by Hellstorm                                        *
* licensed under the GNU General Public License (GPL) v2           *
* convert.cpp - manages de convert process                         *
*******************************************************************/


#include <libmpeg3.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>


#include <avifile.h>
#include <videoencoder.h>
#include <version.h>
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6)
#include <fourcc.h>
#include <creators.h>
#include <except.h>
#endif

#include "mpeg2.h"

//#define CODEC fccIV50
#define CODEC fccDIV3
//#define CODEC fccMPG4
//#define CODEC fccWMV2

#define AUDIOCODE 0x55

static double t0;
static unsigned char **bufferPtr;
static unsigned char *videoBuffer;
static unsigned char *extraBuffer;
static int width, height, crop, scrop;
static int orig_width, orig_height;
static int hasAudio = 1, hasVideo = 1;
static int numFrames = 0;
static int totalnFrames = 0;
static int end = 0;
static IAviWriteFile *outputFile = 0;
static IAviVideoWriteStream *outputVideoStream;
static IAviAudioWriteStream *outputAudioStream;
int numSamplesAudioBuffer;
float numExtraSamples;
float samplesBufferError;
int frameTime;
char *audioBuffer;
int achan;
CImage *im;
BitmapInfo *bh;

void convertSigHandler(int sig)
{
	printf("Signal catched\n");
	fflush(stdout);
	end = 1;
}

void convertInitOutput(char *outputFileName)
{
	try {
		outputFile = CreateIAviWriteFile(outputFileName);
	} catch(FatalError & e) {
		e.Print();
		exit(1);
	}
}

void convertInitVideo(mpeg3_t * file, int stream, float ratio, int bitrate,
		      int dwidth, int dheight, int dcrop, double cropRatio,
		      int crisp)
{
	int i;

	if (mpeg2NumVideoStreams(file) == 0) {
		hasVideo = 0;
		return;
	}

	orig_width = mpeg2VideoWidth(file, stream);
	orig_height = mpeg2VideoHeight(file, stream);

	if (dwidth == -1)
		width = (int) (orig_height * ratio + .5);
	else if (dwidth == 0)
		width = orig_width;
	else
		width = dwidth;
	if (dheight == -1)
		height = orig_height;
	else if (dheight == 0)
		height = (int) (orig_width / ratio + .5);
	else
		height = dheight;


	if (cropRatio != -1) {
		double tmpCrop = (height - width / cropRatio) / 2.0;
		crop = (int) (tmpCrop + 0.5);
		scrop =
		    (int) ((double) crop / (double) height *
			   (double) orig_height + .5);
		height -= (int) (tmpCrop * 2 + 0.5);
		printf
		    ("Ratio based cropping. source lines:%d, dest. lines:%d\n",
		     scrop, crop);
	} else if (dcrop != -1) {
		crop = dcrop;
		scrop =
		    (int) ((double) crop / (double) height *
			   (double) orig_height + .5);
		height -= crop * 2;
		printf
		    ("Lines based cropping. source lines:%d, dest. lines:%d\n",
		     scrop, crop);
	}

	printf("Destination size: width:%d height:%d\n", width, height);

	videoBuffer = new unsigned char[width * height * 3 + 64];
	if (videoBuffer == NULL) {
		fprintf(stderr,
			"Out of memory at convertInitVideo creating buffer\n");
		exit(1);
	}
	bufferPtr = new unsigned char *[height + 2];
	if (bufferPtr == NULL) {
		fprintf(stderr,
			"Out of memory at playInitVideo creating bufferPtr\n");
		exit(1);
	}
	//Idea for this taken from x2divx by Ulrich Hecht -  uli@emulinks.de
	extraBuffer = new unsigned char[width * 3 + 64];
	if (extraBuffer == NULL) {
		fprintf(stderr,
			"Out of memory at playInitVideo creating bufferPtr\n");
		exit(1);
	}
	for (i = 0; i < height; i++) {
		bufferPtr[i] = &videoBuffer[(height - 1 - i) * width * 3];
	}
	bufferPtr[height] = bufferPtr[height - 1];
	bufferPtr[height - 1] = extraBuffer;

	//Setup output
	frameTime = (int) (1000000 / mpeg2VideoFrameRate(file, stream));
	bh = new BitmapInfo(width, height, 24);
	try {
#if (AVIFILE_MAJOR_VERSION == 0) && (AVIFILE_MINOR_VERSION==6)
		IVideoEncoder *enc =
		    Creators::CreateVideoEncoder(CODEC, *bh);
		Creators::SetCodecAttr(enc->GetCodecInfo(), "Crispness",
				       crisp);
		Creators::SetCodecAttr(enc->GetCodecInfo(), "BitRate",
				       bitrate);
		Creators::SetCodecAttr(enc->GetCodecInfo(), "KeyFrames",
				       25);
#else
		IVideoEncoder::SetExtendedAttr(CODEC, "BitRate", bitrate);
		IVideoEncoder::SetExtendedAttr(CODEC, "Crispness", crisp);
		IVideoEncoder::SetExtendedAttr(CODEC, "KeyFrames", 25);
#endif
		outputVideoStream =
		    outputFile->AddVideoStream(CODEC, bh, frameTime);
		outputVideoStream->SetKeyFrame(100);
		outputVideoStream->Start();
		im = new CImage(bh, videoBuffer, false);
	} catch(FatalError & e) {
		e.Print();
		exit(1);
	}
}

void convertInitAudio(mpeg3_t * file, int stream, int bitrate)
{
	if (mpeg2NumAudioStreams(file) == 0) {
		hasAudio = 0;
		return;
	}

	achan = mpeg2AudioNumChannels(file, stream);
	int rate = mpeg2AudioSampleRate(file, stream);

	samplesBufferError =
	    (float) rate / (float) mpeg2VideoFrameRate(file, stream);
	numSamplesAudioBuffer = (int) samplesBufferError;
	samplesBufferError -= numSamplesAudioBuffer;
	numExtraSamples = 0;

	audioBuffer = new char[numSamplesAudioBuffer * 2 * achan];
	if (!audioBuffer) {
		fprintf(stderr,
			"Out of memory at convertInitAudio creating buffer\n");
		exit(1);
	}

	WAVEFORMATEX fmt;
	memset(&fmt, 0, sizeof(WAVEFORMATEX));
	fmt.wFormatTag = AUDIOCODE;
	fmt.nChannels = (achan > 2 ? 2 : achan);
	fmt.nSamplesPerSec = rate;
	fmt.nAvgBytesPerSec = (achan > 1 ? rate : rate * 2);
	fmt.nBlockAlign = (achan > 1 ? 2 : 1);
	fmt.wBitsPerSample = 16;
	fmt.cbSize = sizeof(WAVEFORMATEX);

	try {
		outputAudioStream =
		    outputFile->AddAudioStream(AUDIOCODE, &fmt,
					       (bitrate * 1000) / 8);
		outputAudioStream->SetQuality(20000);
		outputAudioStream->Start();
	}
	catch(FatalError & e) {
		e.Print();
		exit(1);
	}
}

static void wraudio(char *buf, int n, double audiocorrection)
{
	int i = n;
	short int *s = (short int *) buf;
	do {
		*s = (int) (*s * audiocorrection + .5);
		s++;
	} while (--i);
	outputAudioStream->AddData(buf, n);
}



void convertFile(mpeg3_t * file, int astream, int vstream, float ratio,
		 char *outputFileName, int bitrate, int mp3bitrate,
		 int dwidth, int dheight, int crop, double cropRatio,
		 int crisp, int filenr, int lastfile,
		 double audiocorrection)
{
	long numSamples = 0;
	long numVideoSec;
	struct timeval tv;
	double t1;
	int i;

	if (end)
		return;
	if (filenr == 1) {
		hasAudio = 1;
		hasVideo = 1;
		convertInitOutput(outputFileName);
		convertInitVideo(file, vstream, ratio, bitrate, dwidth,
				 dheight, crop, cropRatio, crisp);
		convertInitAudio(file, vstream, mp3bitrate);

		gettimeofday(&tv, NULL);
		t0 = (double) tv.tv_sec * 1000000.0 + (double) tv.tv_usec;
		end = 0;
		signal(SIGINT, convertSigHandler);
	}

	numFrames = 0;
	float frameRate = mpeg2VideoFrameRate(file, vstream);
	int extraBufferLen = width * 3;
	int totalFrames = 0;

	while (!mpeg3_end_of_video(file, vstream) && !end) {
		i = mpeg2VideoReadFrame(file, bufferPtr, 0, scrop,
					orig_width,
					orig_height - scrop - scrop, width,
					height, 3, vstream);
		if (i) {
			printf("\nSkipping frame\n");
			fflush(stdout);
			continue;
		}
		bcopy(bufferPtr[height - 1], bufferPtr[height],
		      extraBufferLen);
		if (hasAudio) {
			mpeg2AudioRead(file, audioBuffer,
				       numSamplesAudioBuffer, astream);
			if (achan == 1)
				wraudio(audioBuffer, numSamplesAudioBuffer,
					audiocorrection);
			else
				wraudio(audioBuffer,
					numSamplesAudioBuffer * 2,
					audiocorrection);
			numExtraSamples += samplesBufferError;
			if (numExtraSamples > 1) {
				mpeg2AudioRead(file, audioBuffer,
					       1, astream);
				if (achan == 1)
					wraudio(audioBuffer, 1,
						audiocorrection);
				else
					wraudio(audioBuffer, 2,
						audiocorrection);
				numExtraSamples -= 1;
			}
		}
		outputVideoStream->AddFrame(im);
		numFrames++;
		printf("Num Frames:%d\r", numFrames);
		fflush(stdout);
	}

	if (lastfile) {
		printf("Closing output file\n");
		fflush(stdout);
		delete outputFile;
	}

	totalnFrames += numFrames;
	gettimeofday(&tv, NULL);
	t1 = (double) tv.tv_sec * 1000000.0 + (double) tv.tv_usec;

	printf("\nReal time elapsed:%.10f\n", (t1 - t0) / 1000000.0);
	printf("Total frames:%d\n", totalnFrames);
	printf("FPS:%.10f\n", totalnFrames / ((t1 - t0) / 1000000.0));
	printf("Decoded %d min. %02d sec.\n",
	       (int) (totalnFrames / frameRate / 60.0),
	       (int) ((int) (totalnFrames / frameRate) % 60));
}


static void countaudio(char *buf, int n, double *ab)
{
	int i = n;
	short int *s = (short int *) buf;
	char c;
	do {
		c = (char) (*s++ >> 8);
		if (c < 0)
			c = -c;
		ab[c] += 1.;
	}
	while (--i);
}


int gettime(mpeg3_t * file, int vstream, int astream, int filenr,
	    int lastfile, int *frames, double *audiocorrection)
{
	int result = !0;
	int minutes = 0;
	int i;

	double ab[129];
	double audiomax, audioreal;

	if (end)
		return 0;

	width = 8;
	height = 8;
	if (filenr == 1) {
		hasAudio = 1;
		numFrames = 0;
		for (i = 0; i < 129; i++)
			ab[i] = 0;
		achan = mpeg2AudioNumChannels(file, astream);
		int rate = mpeg2AudioSampleRate(file, astream);
		printf
		    ("\n# of audio channels: %d\naudio sample rate: %d\n",
		     achan, rate);

		samplesBufferError =
		    (float) rate / (float) mpeg2VideoFrameRate(file,
							       vstream);
		numSamplesAudioBuffer = (int) samplesBufferError;
		samplesBufferError -= numSamplesAudioBuffer;
		numExtraSamples = 0;
		printf("audio samples per frame: %d\n",
		       numSamplesAudioBuffer);

		audioBuffer = new char[numSamplesAudioBuffer * 2 * achan];
		if (audioBuffer == NULL)
			return 0;
	}
	printf("\ncalculating audio correction...\n");
	while (!mpeg3_end_of_audio(file, astream) && !end) {
		mpeg2AudioRead(file, audioBuffer, numSamplesAudioBuffer,
			       astream);
		if (achan == 1)
			countaudio(audioBuffer, numSamplesAudioBuffer, ab);
		else
			countaudio(audioBuffer, numSamplesAudioBuffer * 2,
				   ab);
		numExtraSamples += samplesBufferError;
		if (numExtraSamples > 1) {
			mpeg2AudioRead(file, audioBuffer, 1, astream);
			if (achan == 1)
				countaudio(audioBuffer, 1, ab);
			else
				countaudio(audioBuffer, 2, ab);
			numExtraSamples -= 1;
		}
		numFrames++;
		printf("Num Frames:%d\r", numFrames);
		fflush(stdout);
	}

	if (lastfile) {
		delete audioBuffer;
		*frames = numFrames;

		audiomax = 0;
		for (i = 0; i < 129; i++)
			audiomax += ab[i];
		audiomax = audiomax * .99995;	// 0.005% of the sound may be disformed
		audioreal = 0;
		for (i = 0; i < 129; i++) {
			audioreal += ab[i];
			if (audioreal >= audiomax)
				break;
		}

		*audiocorrection = 128. / (i + 1.);
		if (*audiocorrection < 1.)
			*audiocorrection = 1;
		printf("\naudio correction: %f\n", *audiocorrection);
	}

	return result;
}
