/*******************************************************************
* DivXripper - by Hellstorm                                        *
* licensed under the GNU General Public License (GPL) v2           *
* mpeg2.cpp - interface with libmpeg3                              *
*******************************************************************/

#include <libmpeg3.h>

int mpeg2IsFileCompatible(char *fileName)
{
	return mpeg3_check_sig(fileName);
}

mpeg3_t *mpeg2OpenFile(char *fileName)
{
	mpeg3_t *i;
	i = mpeg3_open(fileName);
/*
	if(i)
	{
		mpeg3_set_mmx(i, !0);
//		mpeg3_seek_percentage(i, 0.0);
	}
*/
	return i;
}

mpeg3_t *mpeg2OpenCopyFile(mpeg3_t * file, char *fileName)
{
	mpeg3_t *i;
	i = mpeg3_open_copy(fileName, file);
//      if(i)
//              mpeg3_set_mmx(i, !0);
	return i;
}

void mpeg2CloseFile(mpeg3_t * file)
{
	mpeg3_close(file);
}

void mpeg2UseMMX(mpeg3_t * file, int use_mmx)
{
	mpeg3_set_mmx(file, use_mmx);
}

void mpeg2ShowFileInfo(mpeg3_t * file)
{
	int numAStreams;
	int numVStreams;
	int i;

	if (mpeg3_has_audio(file)) {
		printf("File has audio\n");
		numAStreams = mpeg3_total_astreams(file);
		printf("Number of audio streams: %d\n", numAStreams);
		for (i = 0; i < numAStreams; i++) {
			int numChannels = mpeg3_audio_channels(file, i);
			int sampleRate = mpeg3_sample_rate(file, i);
			int audioSamples = mpeg3_audio_samples(file, i);
			printf("Stream %d:%d channels, %d sample rate,"
			       " %d audioSamples\n", i, numChannels,
			       sampleRate, audioSamples);
		}
	}
	if (mpeg3_has_video(file)) {
		printf("File has video\n");
		numVStreams = mpeg3_total_vstreams(file);
		printf("Number of video streams: %d\n", numVStreams);
		for (i = 0; i < numVStreams; i++) {
			int width = mpeg3_video_width(file, i);
			int height = mpeg3_video_height(file, i);
			float frameRate = mpeg3_frame_rate(file, i);
			int videoFrames = mpeg3_video_frames(file, i);
			float aspectRatio = mpeg3_aspect_ratio(file, i);
			printf("Stream %d:%d width, %d height, "
			       "%f frameRate, %d videoFrames,%f aspect ratio\n",
			       i, width, height, frameRate, videoFrames,
			       aspectRatio);
		}
	}

}

int mpeg2VideoGetNumFrames(mpeg3_t * file, int vstream)
{
	return mpeg3_video_frames(file, vstream);
}

int mpeg2AudioNumChannels(mpeg3_t * file, int astream)
{
	return mpeg3_audio_channels(file, astream);
}


int mpeg2AudioSampleRate(mpeg3_t * file, int astream)
{
	return mpeg3_sample_rate(file, astream);
}

float mpeg2VideoFrameRate(mpeg3_t * file, int vstream)
{
	return mpeg3_frame_rate(file, vstream);
}

int mpeg2NumAudioStreams(mpeg3_t * file)
{
	return mpeg3_total_astreams(file);
}

int mpeg2NumVideoStreams(mpeg3_t * file)
{
	return mpeg3_total_vstreams(file);
}

// Taken from xmovie-1.5.4 by Adam Williams broadcast@earthling.net
int mpeg2AudioRead(mpeg3_t * file, char *buffer, long len, int stream)
{
	int channel;
	int i, j;
	int result = 0;
	int16_t *output_ptr;
	int16_t *output_end;
	int16_t *input_ptr;
	int16_t *input_end;
	int total_channels;

	int16_t read_buffer[len];

	if (mpeg3_audio_channels(file, stream) > 2) {
		output_ptr = (int16_t *) buffer;
		output_end = (int16_t *) output_ptr + len * 2;
		while (output_ptr < output_end) {
			*output_ptr++ = 0;
		}
	}

	total_channels = mpeg3_audio_channels(file, stream);

	for (channel = 0; channel < total_channels && !result; channel++) {
		if (channel == 0)
			result = mpeg3_read_audio(file,
						  0,
						  read_buffer,
						  channel, len, stream);
		else
			result = mpeg3_reread_audio(file,
						    0,
						    read_buffer,
						    channel, len, stream);

		if (total_channels == 5) {
			input_ptr = (int16_t *) read_buffer;
			input_end = input_ptr + len;

			switch (channel) {
			case 0:
				output_ptr = (int16_t *) buffer;
				break;
			case 1:
				output_ptr = (int16_t *) buffer;
				break;
			case 2:
				output_ptr = (int16_t *) buffer + 1;
				break;
			case 3:
				output_ptr = (int16_t *) buffer;
				break;
			case 4:
				output_ptr = (int16_t *) buffer + 1;
				break;
			}

			if (channel == 1)
				while (input_ptr < input_end) {
					*output_ptr++ += *input_ptr;
					*output_ptr++ += *input_ptr++;
			} else
				while (input_ptr < input_end) {
					*output_ptr += *input_ptr++;
					output_ptr += 2;
				}
		} else {
			output_ptr = (int16_t *) buffer + channel;
			input_ptr = (int16_t *) read_buffer;
			input_end = input_ptr + len;

			while (input_ptr < input_end) {
				*output_ptr = *input_ptr++;
				if (total_channels == 2)
					output_ptr += 2;
				else
					output_ptr++;
			}
		}
	}
	return result;
}

// We will read the frame with MPEG3_BGR888 format
int mpeg2VideoReadFrame(mpeg3_t * file, unsigned char **buffer, int in_x,
			int in_y, int in_w, int in_h, int out_w, int out_h,
			int depth, int stream)
{
	int mode;

	if (depth == 16)
		mode = MPEG3_RGB565;
	else
		mode = MPEG3_BGR888;
	return mpeg3_read_frame(file, buffer, in_x, in_y, in_w, in_h,
				out_w, out_h, mode, stream);
}

int mpeg2VideoReadYUVFrame(mpeg3_t * file, char *yBuffer, char *uBuffer,
			   char *vBuffer, int in_x, int in_y, int in_w,
			   int in_h, int stream)
{
	return mpeg3_read_yuvframe(file, yBuffer, uBuffer, vBuffer, in_x,
				   in_y, in_w, in_h, stream);
}


int mpeg2VideoWidth(mpeg3_t * file, int stream)
{
	return mpeg3_video_width(file, stream);
}

int mpeg2VideoHeight(mpeg3_t * file, int stream)
{
	return mpeg3_video_height(file, stream);
}
